# Bash Utility Functions related to building RNAstructure Releases.

# Write error message. 
# err  <SOURCE> <MESSAGE> <MSG_ARGS>
err() {
    echo -n >&2  "$1: ERROR: "
    printf  >&2 "$2\n" "${@:3}"
}

abort() {
    local usageError argnum=0 errNum=1
    while (($# > 0)); do
        case "$1" in 
            -U) usageError=1 ;;
            -e[=:]*) errNum=${1:3} ;;
            *) break ;;
        esac
        shift
    done
    err "$SCRIPT_NAME" "$@"
    cd "$PREV_DIR"
    [[ $usageError ]] && showusage
    exit $errNum
}

# readVer() {
#   local verFile=${SRC_DIR:-$DEF_SRC_DIR}/Version.txt
#   #CURVERINFO=$(< $verFile)
#   DEF_VERSION=$(sed -n -E 's/^.*version ([0-9.]+).*$/\1/p' "$verFile")
#   DEF_RELEASE_DATE=$(sed -n -E 's/^.*(released on|date is) ([^.]+).*$/\2/p' "$verFile")
#   #echo "Read VERSION: $DEF_VERSION"
#   #echo "Read RELEASE_DATE: $DEF_RELEASE_DATE"
# }

writeVer() {
    local verFile=${SRC_DIR:-$DEF_SRC_DIR}/Version.txt
    local releaseType=$(toupper $1) # G, T, or S
    local outputDir="$2"
    local pattern='^\w*'${releaseType}'\w*:(.*)'

    [[ -d $outputDir ]] || abort 'Failed to write version information. Source directory is missing or invalid: %s' "$outputDir"

    # Write Version.txt
    sed -E -n "s/$pattern/\\1/p" "$verFile"  |  replaceVars  > "$outputDir/Version.txt" || abort 'Failed to write version information to %s' "$outputFile"
    
    # Write src/version.h (but only if the src directory already exists. Otherwise, version.h may not be needed.)
    [[ ! -d $outputDir/src ]] || replaceVars < "$RES_DIR/version.h.tpl" > "$outputDir/src/version.h"  || abort 'Failed to write version information to %s' "$outputDir/src/version.h"
}

replaceVars() {
    local VER="$VERSION" DATE="$RELEASE_DATE"
    local TODAY=$(date '+%B %e, %Y')
    local vars=(VERSION VER RELEASE_DATE DATE ARCH OS BITS TODAY JDK_DIR)
    
    if (($#==0)); then
        local text=$(cat)
    else
        local text="$*"
    fi
    
    for var in ${vars[@]}; do
        if [[ $text == *\[$var\]* ]]; then
            val="${!var}"
            #echo replace $var with "$val"
            text="${text//\[$var\]/$val}"
        #else
            #echo no $var
        fi
    done
    echo -n "$text"
}

# Scripted file and directory copying
# copyEx <INITIAL_SRC_DIR> <INITIAL_DEST_DIR> [<DIRECTIVE>|<ITEM>] ...
# A <DIRECTIVE> is a command to change the current source directory (SRC_DIR) or current destination directory (DST_DIR) or both.
# An <ITEM> is a file or directory to copy, optionally specifying the destination file or directory name.
# DIRECTIVES:
#   --s:PATH  -- Runs 'cd PATH' (which changes SRC_DIR relative to its current value.)
#   --d:PATH  -- Appends PATH to the current destination directory (DST_DIR).
#   --c:PATH  -- Same as --s:PATH  --d:PATH  (i.e. changes both souce and destination directories)
#   --S:PATH  -- If PATH is relative, the current source directory (SRC_DIR) is set to INITIAL_SRC_DIR/PATH. Otherwise SRC_DIR is set to the absolute PATH specified.
#   --D:PATH  -- If PATH is relative, the current destination directory (DST_DIR) is set INITIAL_DEST_DIR/PATH. Otherwise, DST_DIR is set to the absolute PATH specified.
#   --C:PATH  -- Same as --S:PATH  --D:PATH  (i.e. sets both souce and destination directories)
#   --e:x|i   -- Sets the action to take if a copy operation fails:
#                   x -- (default) Exit with an error code indicating the number of the instruction that failed (starting with 1)
#                   i -- Ignore the error and continue copying
#                Note that DIRECTIVES (such as 'cd') will ALWAYS cause copyEx to exit immediately if they fail, regardless of --e.
#   --g:x|i   -- Sets the action to take if a glob PATTERN fails to match any files. (See --e for the meaning of x and i).
#   --i       -- Ignore errors in the following single ITEM, DIRECTIVE, or PATTERN (when --e:x)
#   --x       -- Exit if there is an error in the following single ITEM, DIRECTIVE, or PATTERN  (when --e:i)
# ITEMS:
#   PATTERN         -- Copy all files and directories that match PATTERN from SRC_DIR to DST_DIR.
#   PATTERN=DIR     -- Copies all files and directories that match PATTERN from SRC_DIR into the directory DST_DIR/DIR/
#   ITEM==NEW_NAME  -- Copies SRC_DIR/ITEM into DST_DIR with the new name NEW_NAME (ITEM can be a file or directory, and both ITEM and NEW_NAME can include sub-directory parts e.g. 'sub/dir/item==some/other/new_item')
copyEx() {(   
    # NOTE this runs inside a subshell to preserve the outer CWD and environment.
    local initSrcDir="$1" initDestDir="$2"; shift 2
    local arg argNum cmd val errMode tmpErrMode newTmpErrMode globMode srcDir dstDir item dstItem

    cd "$initSrcDir" && srcDir=$PWD || { err 'Failed to enter initial source directory: "%s"' "$initSrcDir";   exit 101; }
    initSrcDir="$srcDir"

    # initDestDir can be relative or absolute, but it may not exist. We need to make it absolute, so that changes in cd don't affect it
    initDestDir=$(normpath "$initDestDir")
    dstDir="$initDestDir"

# [[ $dstDir ]] || { err 'Failed to create or enter initial destination directory: "%s"' "$initDestDir";   exit 101; }

    for arg in "$@"; do
        ((argNum++))
        [[ -z $arg ]] && continue;
        case "$arg" in
            --?:*) # It is a DIRECTIVE
                val="${arg:5}"
                case ${arg:3:1} in
                    s)  cd "$val"
                        ;;
                    d)  dstDir=$(cleanpath "$dstDir/$val")
                        ;;
                    c)  cd "$val"
                        dstDir=$(cleanpath "$dstDir/$val")
                        ;;
                    S)  cd "$initSrcDir" && cd "$val" || { [[ $tmpErrMode == i ]] || exit $argNum; }
                        ;;
                    D)  dstDir=$(normpath "$val" --base="$initDestDir")
                        ;;
                    C)  cd "$initSrcDir" && cd "$val" || { [[ $tmpErrMode == i ]] || exit $argNum; }
                        dstDir=$(normpath "$val" --base="$initDestDir")
                        ;;
                    e)  [[ $val == [xi] ]] && errMode=$val  || { [[ $tmpErrMode == i ]] || exit $argNum; }
                        ;;
                    g)  [[ $val == [xi] ]] && globMode=$val || { [[ $tmpErrMode == i ]] || exit $argNum; }
                        ;;
                esac
                ;;
            --[ix]) # It is a DIRECTIVE (--i or --x)
                tmpErrMode=${arg:3:1}
                ;;
            ?*=?*)
                #newDir=${item#*=}
                #cd $newDir
                #$destDir+=/$newDir
                ;;
        esac
        tmpErrMode= #clear this temporary value.
    done
)}

buildTar() {  #  buildTar outputFile [TAR_FLAGS...] [FILES...]
    local outputFile="$1"; shift
    local outputDir=$(dirname "$outputFile")
    #local excludeFile= # write each item from the array to a temporary file buffer
    [[ -d $outputDir ]] || mkdir -p "$outputDir" || abort 'buildTar: Failed to create output directory for tar creation: "%s"' "$outputDir"
    #(IFS=$'\n'; echo "${TAR_EXCLUDE[*]}" > "$excludeFile")  # Write current list of exclusion patterns to the file (one on each line)
    local flags=" "  # Flags that are desirable, but are Not supported by all versions: --exclude-vcs-ignores  --owner=$TAR_OWNER --group=$TAR_GROUP  --exclude-backups --exclude-vcs 
    # local renameRootFlags='--transform=|\./\?|/RNAstructure/'
    tar -zvcf $outputFile $flags \
        --exclude-from=<(printf '%s\n' "${TAR_EXCLUDE[@]}") \
        "$@" || 
        abort 'buildTar: Failed to create tar: "%s"' "$outputFile"
}


## cpsync <SOURCE> <DEST> [rsync_FLAGS...]
## Copies files from source to destination without overwriting any newly created files. 
cpsync() {
    rsync -au --exclude='.git/' --exclude-from=<(printf '%s\n' "${SYNC_EXCLUDE[@]}") "$@" | 
        grep -E -v '/CVS/|^CVS/|/$' # remove output from CVS folders (and their contents) as well as other lines listing directories (i.e. lines ending in /)
    ((${PIPESTATUS[0]}==0)) ||
        abort 'cpsync: Failed to copy source files from (%s) to "%s".' "${*:1:$#-1}" "${@: -1}"
}

## usage: parseTasks [t1 t2...] | [+/-t1 t2...]...        
parseTasks() {
    local tasks=( ${1//,/ } )         #replace commas with spaces and split task lists
    local taskFlags=( ${2//,/ } )     #replace commas with spaces and split task lists
    # echo tasks before: $(joinArgs '|' "${tasks[@]}")  "(${#tasks[@]})"
    # echo taskFlags before: $(joinArgs '|' "${taskFlags[@]}") "(${#taskFlags[@]})"
    if [[ ${#taskFlags[@]} -ne 0 ]]; then 
        if [[ ${taskFlags[0]} == [+-]* ]]; then
            # taskFlags corresponds to additions (+) and removals (-)
            local taskMode task
            for task in "${taskFlags[@]}"; do
                [[ $task == [+-]* ]] && { taskMode=${task:0:1}; task=${task:1}; }
                [[ $task ]] || continue
                if [[ $taskMode == '+' ]]; then
                    tasks+=($task)
                elif [[ $taskMode == '-' ]]; then
                    tasks=( ${tasks[@]/$task/} )
                fi
            done
        else
            # taskFlags corresponds to an exact set
            tasks=( ${taskFlags[@]} )
        fi
    fi
    local IFS=','
    echo "${tasks[*]}"
    unset IFS
    # echo tasks after: $tasks; exit
}