/*
 * (c) 2011 Mathews Lab, University of Rochester Medical Center.
 *
 * This software is part of a group specifically designed for the RNAstructure
 * secondary structure prediction program.
 */

package ur_rna.RNAstructureUI.menus;

import ur_rna.RNAstructureUI.AppMainFrame;
import ur_rna.RNAstructureUI.RNAstructure;
import ur_rna.RNAstructureUI.utilities.Dialogs;
import ur_rna.Utilities.AppLog;
import ur_rna.Utilities.PathTools;

import java.awt.*;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.net.URI;

import static ur_rna.Utilities.ObjTools.asBool;
import static ur_rna.Utilities.ObjTools.isEmpty;

/**
 * A class that creates a "Help" menu.
 *
 * @author Jessica S. Reuter
 */
public class HelpMenu
	extends RolloverMenu {
	private static final long serialVersionUID = 20120802;
	private static final String WEB_URL = "http://rna.urmc.rochester.edu/";
	private static final String HELP_URL = WEB_URL + "GUI/html/Contents.html";
	private static final String LOCAL_HELP_PAGE = "Contents.html";

	/**
	 * Constructor.
	 */
	public HelpMenu() {
		super( "Help" );
		addMenuItem( "Program Help", "View the RNAstructure Manual.", "F1" );
		addMenuItem( "Online Help", "View the online RNAstructure help.", "*F1");

		if (PathTools.findLocalPath("examples") != null) {
			addSeparator();
			addMenuItem("Extract Examples", "Extract example files to a folder of your choice.");
		}
		addSeparator();

		if (asBool(System.getProperty("show-gui-info-menu"))) {
			addMenuItem("Show GUI Info", "Output information about the GUI Component hierarchy to stdout.");
			addSeparator();
		}

		addMenuItem( "About RNAstructure...",
			"Display program information, version number, and copyright." );
	}

	@Override
	protected void doMenuActions( String command ) {
		// If the command is to show the about screen, do so.
		if( command.startsWith( "About" ) ) {
			try {
				//Dialogs.showMessage( ImageGrabber.getImageLabel( "logo.png" ) );
				AppMainFrame.getFrame().showAboutWindow();
			} catch( Exception e ) {
				Dialogs.showError( "error showing about screen.\n" + AppLog.getErrorInfo(e));
			}
		}

		// Otherwise, show the help in a browser window.
		else if ("Program Help".equals(command)){
			String uri = HELP_URL;
			File page = PathTools.findLocalPath("manual/html/" + LOCAL_HELP_PAGE);
			if (page == null) page = PathTools.findLocalPath("manual/GUI/html/" + LOCAL_HELP_PAGE);
			if (page != null)
				uri = page.toURI().toString();
			browse(uri);
		}
		// Otherwise, show the help in a browser window.
		else if ("Online Help".equals(command)){
			browseHelp();
		}
		else if ("Extract Examples".equals(command)) {
			RNAstructure.ExtractExamples();
		}
		else if ("Show GUI Info".equals(command)) {
			for (Window w: Window.getWindows())
				addGuiInfo(w);
		}
	}
	private void addGuiInfo(final Component c) {
		if (c instanceof Window) {
			Window w = (Window)c;
			for (Window ow : w.getOwnedWindows())
				addGuiInfo(ow);
		}
		c.removeMouseListener(guiInfoListener);
		c.addMouseListener(guiInfoListener);
		if (c instanceof Container) {
			for (Component co : ((Container) c).getComponents())
				addGuiInfo(co);
		}
	}
	private final MouseAdapter guiInfoListener = new MouseAdapter() {
		@Override
		public void mouseClicked(final MouseEvent e) {
			AppMainFrame.getFrame().setInfoLabel(getGuiInfo(e.getSource()));
		}
	};
	private String getGuiInfo(Object source) {
		if (source == null)
			return "Source: NULL";
		if (!(source instanceof Component))
			return "Source Type: " + source.getClass().getSimpleName();
		Component c = (Component)source;
		if (c.getParent() == null)
			return String.format("%s '%s' :: [TOP]", c.getClass().getSimpleName(), c.getName());
		return String.format("%s '%s' :: %s '%s'", c.getClass().getSimpleName(), c.getName(), c.getParent().getClass().getSimpleName(), c.getParent().getName());
	}
	public static void browseHelp() {
		browse(HELP_URL);
	}
	public static void browseWebsite() {
		browse(WEB_URL);
	}
	public static void browse(String uri) {
		try {
			if (isEmpty(uri)) uri = WEB_URL;
			Desktop.getDesktop().browse(new URI(uri));
		} catch( Exception e ) {
			Dialogs.showError( "error showing webpage in browser: " + uri);
		}
	}
}
