package ur_rna.Utilities;

import ur_rna.Utilities.annotation.MagicConstant;
import ur_rna.Utilities.annotation.NotNull;
import ur_rna.Utilities.annotation.ToFriendlyString;

import java.awt.*;
import java.lang.annotation.Annotation;
import java.lang.reflect.Array;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.*;

public final class ObjTools {
    /**
     * An object that can be used to represent a missing value.
     * For example, a Map's get(key) function will return null under either of two circumstances:
     * (1) the key was not found or (2) the key was associated with a null value.
     * Instead of using null as an ambiguous result, MISSING can be returned in the case of
     * the missing key in such circumstances.
     */
    public static Object MISSING = new Object();

    private ObjTools(){} //prevent instantiation

    public static String fmt(String format, Object... args) {
        return String.format(format, args);
    }

    @SuppressWarnings("unchecked") // getMethod results in a warning about using unchecked or unsafe operations.
    public static String toDisplayString(Object value) {
        if (value == null)
            return "null";
        if (value instanceof String)
            return "\"" + value + "\"";
        if (value instanceof Double || value instanceof Integer)
            return value.toString(); // + value.getClass().getSimpleName().substring(0,1); // D,F,I,B

        Class cls =value.getClass();
        String name = cls.getSimpleName();

        if (value instanceof Number)
            return value.toString() + name.substring(0, 1).toLowerCase(); // D,F,I,B
        if (value instanceof Component) {
            Component c = (Component) value;
            Rectangle r = c.getBounds();
            return "[GUI " + name + " \"" + c.getName() + "\" Bounds: (x:" + r.x + ", y:" + r.y + ", w:" + r.width + ", h:" + r.height + ") ]";
        }
        if (value instanceof Throwable) {
            return "[ERROR: " + ((Throwable) value).getMessage() + "]";
        }
        if (value == MISSING)
            return "[MISSING]";

        if (value instanceof Displayable)
            return ((Displayable) value).toDisplayString();

		/* List classes here that can be converted by using toString() directly, without prepending the class name */
        Class[] knownTypes = new Class[]{Boolean.class};
        for (Class c : knownTypes) {
            if (c.isInstance(value))
                return value.toString();
        }

        Annotation[] notes = cls.getAnnotationsByType(ToFriendlyString.class);
        if (notes.length != 0) {
            ToFriendlyString a = (ToFriendlyString)notes[0];
            String method = a.method();
            if ("toString".equals(method))
                return value.toString();
            if (method != null && !method.isEmpty()) {
                try {
                    Method m = cls.getMethod(method);
                    if (m != null)
                        return m.invoke(value).toString();
                } catch (Exception ex) {
                    // do nothing. use another method.
                }
            }
        }

        String s = value.toString();
        if (s.startsWith("[") && s.endsWith("]") || s.startsWith("{") && s.endsWith("}"))
            return s;
        return "[" + name + ": " + s + "]";
    }
    public static String toStr(Object value) {
        return toStr(value, "");
    }
    public static String toStr(Object value, String nullValue) {
        if (value == null) return nullValue;
        return value.toString();
    }
    /**
     * Used to convert raw characters to their escaped version
     * when these raw version cannot be used as part of a Java
     * string literal.
     * <p>
     * Escapes NULL, BKSP, TAB, CR, LF, FF, backslash, double (") and single (') quotes
     * to their java escape codes. Also escapes all other control characters (1 to 31) and
     * all characters above 127 to the corresponding Java unicode escape code (\u0001, \u26BD etc. )
     * </p>
     *
     * @param subject The string to escape.
     * @return The escaped string.
     * @see #escapeStringLiteral(String,int)
     */
    public static String escapeStringLiteral(String subject) {
        return escapeStringLiteral(subject, CharType.ALL);
    }

    /**
     * Used to convert raw characters to their escaped version
     * when these raw version cannot be used as part of a Java
     * string literal.
     * <p>
     * Escapes NULL, BKSP, TAB, CR, LF, FF, backslash, double (") and single (') quotes
     * to their java escape codes. Also escapes all other control characters (1 to 31) and
     * all characters above 127 to the corresponding Java unicode escape code (\u0001, \u26BD etc. )
     * </p>
     *
     * @param subject The string to escape.
     *
     * @param escapeOptions
     *   A bitfield of {@link CharType} values indicating which types
     *   of characters to escape.
     *   <p>
     *   However note {@link CharType#CONTROL} characters are ALWAYS escaped, while
     *   {@link CharType#ALPHA}, {@link CharType#DIGITS}, and most {@link CharType#SYMBOLS}
     *   are NEVER escaped.
     *   The only character types for which escaping can be enabled or disabled are the following:
     *   {@link CharType#TAB}, {@link CharType#NEWLINE}, {@link CharType#DOUBLE_QUOTES},
     *   {@link CharType#SINGLE_QUOTE}, {@link CharType#BACKSLASH}, and {@link CharType#EXT_ASCII}.
     *   </p>
     *
     * @return The escaped string.
     * @see #escapeStringLiteral(String, int)
     */
    public static String escapeStringLiteral(String subject,
            @MagicConstant(flagsFromClass = CharType.class) int escapeOptions) {
        boolean bDQ = isSet(CharType.DOUBLE_QUOTES, escapeOptions);
        boolean bSQ = isSet(CharType.SINGLE_QUOTE, escapeOptions);
        boolean bEA = isSet(CharType.EXT_ASCII, escapeOptions);
        boolean bSL = isSet(CharType.SLASH, escapeOptions);

        StringBuilder sb = new StringBuilder(subject.length());
        char c;
        for (int i = 0; i < subject.length(); i++) {
            switch (c = subject.charAt(i)) {
                case 0:
                    sb.append("\\0");
                    continue;
                case CharConstants.BKSP:
                    sb.append("\\b");
                    continue;
                case CharConstants.TAB:
                    sb.append("\\t");
                    continue;
                case CharConstants.LF:
                    sb.append("\\n");
                    continue;
                case CharConstants.FF:
                    sb.append("\\f");
                    continue;
                case CharConstants.CR:
                    sb.append("\\r");
                    continue;
                case CharConstants.DBL_QUOT:
                    sb.append(bDQ ? "\\\"" : c);
                    continue;
                case CharConstants.SNG_QUOT:
                    sb.append(bSQ ? "\\\'" : c);
                    continue;
                case CharConstants.BACK_SLASH:
                    sb.append(bSL ? "\\\\" : c);
                    continue;
                case CharConstants.DEL:
                    sb.append("\\u007f");
                default:
                    if (c < CharConstants.SP || bEA && c >= CharConstants.UPPER_ASCII || c > CharConstants.LAST_ASCII) {
                        String s = "000" + Integer.toString(c, Constants.HEX_RADIX);
                        sb.append("\\u");
                        sb.append(s, s.length() - 4, s.length());
                    } else {
                        sb.append(c);
                    }
            }
        }
        return sb.toString();
    }

    public static boolean isSet(final int flagToFind, final int valueToSearch) {
        return (valueToSearch & flagToFind) == flagToFind;
    }
    public static boolean isAnySet(final int flagToFind, final int valueToSearch) {
        return (valueToSearch & flagToFind) != 0;
    }
    /**
     * Returns true if the string is null or empty ("").
     * @param str
     * @return
     */
    public static boolean isEmpty(final String str) {
        return str == null || str.isEmpty();
    }

    /**
     * If obj is a boolean, its value is returned.
     * If obj is null, an empty array, or a number, char or byte representing 0, the return value is False.
     * If obj is a string, then it is evaluated as described in AsBool(string)
     * Otherwise the return value is the the result of AsBool(string) applied to obj.ToString()
     */    
    public static boolean asBool(final Object obj) {
        if (obj == null) return false;
        if (obj instanceof String) return asBool((String) obj, false);
        if (obj instanceof Boolean)
            return (Boolean) obj;
        if (obj instanceof Byte)
            return 0 != (byte) obj;
        if (obj instanceof Integer)
            return 0 != (int) obj;
        if (obj instanceof Float)
            return 0F != (float) obj;
        if (obj instanceof Long)
            return 0L != (long) obj;
        if (obj instanceof Character)
            return 0 == (Character)obj;
        if (obj instanceof Array)
            return 0 != Array.getLength(obj);
        return asBool(obj.toString()); //uses string version. "", false, no, "0" evaluate to false. All others are true, including "{TypeName}" etc.
    }

    /**
    * Attempts to convert a string representing a conceptual "true/false" into a boolean.
    * @return  {@code defaultIfEmpty } if {@code boolString} is null, empty (""), or whitespace.
    *           Returns FALSE if {@code boolString} == "false", "no" or "0" (case insensitive, trimmed).
    *           True otherwise.
    */
    public static boolean asBool(String boolString, boolean defaultIfEmpty) {
        if (boolString == null) return defaultIfEmpty;
        boolString = boolString.trim().toLowerCase();
        if ("".equals(boolString)) return defaultIfEmpty; //check length again AFTER trim()
        switch (boolString) {
            case "false":
            case "0":
            case "no":
                return false;
        }
        return true; //anything else, return true.
    }

    public static <T> void copyTo(java.util.List<T> list, T[] arr) { copyTo(list, arr, 0, list.size()); }
    public static <T> void copyTo(java.util.List<T> list, T[] arr, int start) { copyTo(list, arr, start, list.size()); }
    public static <T> void copyTo(java.util.List<T> list, T[] arr, int start, int length) {
        for (int i = 0; i < length; i++)
            arr[start + i] = list.get(i);
    }

    @SuppressWarnings("unchecked")
    public static <T> ArrayList<T> safeClone(ArrayList<T> list) {
        return (ArrayList<T>)list.clone();
    }
    private static java.lang.reflect.Method cloneMethod = getCloneMethod();
    private static Method getCloneMethod() {
        try {
            return Object.class.getDeclaredMethod("clone");
        }catch (NoSuchMethodException ex) {
            for (java.lang.reflect.Method m : Object.class.getDeclaredMethods())
                AppLog.getDefault().warn("Found Method: " + m.getName());
            //should never happen because Object has a method called clone.
            throw new InternalError(ex);
        }
    }
    @SuppressWarnings("unchecked")
    public static <T> T safeClone(T obj) {
        try {
            return (T) cloneMethod.invoke(obj);
        } catch (IllegalAccessException | InvocationTargetException ex) {
            return  null;
        }
    }
    @SuppressWarnings("unchecked")
    public static <T> T firstOfType(final Object[] arr, final Class<T> type) {
        Objects.requireNonNull(type);
        Objects.requireNonNull(arr);
        for(Object item : arr) {
            if (item != null && type.isInstance(item))
                return (T)item;
        }
        return null;
    }

    /**
     * Returns the first string in the list of parameters that is not null or empty ("")
     * Returns empty ("") if all strings are either null or empty.
     */
    @NotNull
    public static String firstNonEmpty(final String... names) {
        for (String name : names) {
            if (!isEmpty(name))
                return name;
        }
        return "";
    }

    public static String escapeHTML(String s) {
        StringBuilder out = new StringBuilder(Math.max(16, s.length()));
        for (int i = 0; i < s.length(); i++) {
            char c = s.charAt(i);
            switch (c) {
                //case '"': out.append("&quot;"); break;
                case '<': out.append("&lt;"); break;
                case '>': out.append("&gt;"); break;
                case '&': out.append("&amp;"); break;
                default:
//                    if (c >= 0x7F) {
//                        out.append("&#");
//                        out.append((int) c);
//                        out.append(';');
//                    } else
                        out.append(c);
            }
        }
        return out.toString();
    }

    public static <T> int indexOf(final T[] subject, final T find) {
        if (find == null) {
            for (int i = 0; i < subject.length; i++)
                if (subject[i] == null) return i;
        } else {
            for (int i = 0; i < subject.length; i++)
                if (find.equals(subject[i]))
                    return i;
        }
        return -1;
    }

    public static int indexOf(final String[] subject, final String find) { return indexOf(subject, find, false); }
    public static int indexOf(final String[] subject, final String find, boolean ignoreCase) {
        if (find == null) {
            for (int i = 0; i < subject.length; i++)
                if (subject[i] == null) return i;
        } else if (ignoreCase) {
            for (int i = 0; i < subject.length; i++)
                if (find.equalsIgnoreCase(subject[i]))
                    return i;
        } else {
            for (int i = 0; i < subject.length; i++)
                if (find.equals(subject[i]))
                    return i;
        }
        return -1;
    }
}
